__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from pathlib import Path
from typing import Any, Callable, Dict, List, Tuple

AlgorithmOptions = Dict[str, float]
ComparisonFunction = Callable[[Path, Path, AlgorithmOptions], bool]
ActionFunction = Callable[[Tuple[Path, Path]], Any]
Results = List[Tuple[Path, Path]]



from math import sqrt, fabs
from typing import Iterator, List

from duplicate_images.image_wrapper import ImageWrapper, get_aspect_ratio


def aspects_roughly_equal(
        image: ImageWrapper, other_image: ImageWrapper, aspect_fuzziness: float
) -> float:
    return fabs(get_aspect_ratio(image, other_image) - 1) < aspect_fuzziness


def compare_image_histograms(
        image: ImageWrapper, other_image: ImageWrapper, aspect_fuzziness: float, rms_error: float
) -> bool:
    def get_deviations(hist: List[float], other_hist: List[float]) -> Iterator[float]:
        return map(lambda a, b: (a - b) ** 2, hist, other_hist)

    if not image.valid or not other_image.valid:
        return False
    if not aspects_roughly_equal(image, other_image, aspect_fuzziness):
        return False
    deviations = get_deviations(image.get_histogram(), other_image.get_histogram())
    rms = sqrt(sum(deviations) / len(image.get_histogram()))
    return rms < rms_error



#!/usr/bin/env /usr/bin/python3

import logging
from dataclasses import dataclass
from functools import partial
from multiprocessing.dummy import Pool
from os import walk
from pathlib import Path
from typing import Callable, List, Optional

from duplicate_images.function_types import ComparisonFunction, AlgorithmOptions, Results
from duplicate_images.image_wrapper import ImageWrapper
from duplicate_images.methods import COMPARISON_METHODS, ACTIONS_ON_EQUALITY
from duplicate_images.parse_commandline import parse_command_line

CHUNK_SIZE = 25


@dataclass
class ParallelOptions:
    parallel: bool = False
    chunk_size: int = CHUNK_SIZE


def files_in_dirs(
        dir_names: List[Path], is_file: Callable[[Path], bool] = lambda f: f.is_file()
) -> List[Path]:
    """Returns a list of all files in directory dir_name, recursively scanning subdirectories"""
    files = [
        Path(root) / filename
        for dir_name in dir_names
        for root, _, filenames in walk(dir_name)
        for filename in filenames
        if is_file(Path(root) / filename)
    ]
    return files


def pool_filter(
        candidates: Results,
        compare_images: ComparisonFunction,
        options: AlgorithmOptions, chunk_size: float
) -> Results:
    pool = Pool(None)
    to_keep = pool.starmap(
        partial(compare_images, options=options),
        candidates, chunksize=chunk_size
    )
    return [c for c, keep in zip(candidates, to_keep) if keep]


def similar_images(
        files: List[Path], compare_images: ComparisonFunction,
        options: AlgorithmOptions, parallel_options: ParallelOptions = ParallelOptions()
) -> Results:
    """Returns all pairs of image files in the list files that are exactly_equal
       according to comparison function compare_images"""
    if parallel_options.parallel:
        candidates = [
            (file, other_file)
            for file in files
            for other_file in files[files.index(file) + 1:]
        ]
        return pool_filter(
            candidates, compare_images, options, parallel_options.chunk_size
        )

    return [
        (file, other_file)
        for file in files
        for other_file in files[files.index(file) + 1:]
        if compare_images(file, other_file, options)
    ]


def get_matches(
        root_directories: List[Path], algorithm: str,
        options: Optional[AlgorithmOptions] = None,
        parallel_options: ParallelOptions = ParallelOptions()
) -> Results:
    options = {} if options is None else options
    comparison_function = COMPARISON_METHODS[algorithm]
    image_files = sorted(files_in_dirs(root_directories, ImageWrapper.is_image_file))
    logging.info("%d total files", len(image_files))

    matches = similar_images(
        image_files, comparison_function,
        options, parallel_options
    )
    return matches


def execute_actions(matches: Results, action_name: str) -> None:
    action_equal = ACTIONS_ON_EQUALITY[action_name]
    for match in sorted(matches):
        try:
            action_equal(match)
        except FileNotFoundError:
            continue


def main() -> None:
    args = parse_command_line()
    log_level = logging.DEBUG if args.debug else logging.INFO
    logging.basicConfig(format='%(levelname)s: %(message)s', level=log_level)
    try:
        options = {'aspect_fuzziness': args.aspect_fuzziness, 'rms_error': args.fuzziness}
        matches = get_matches(
            [Path(folder) for folder in args.root_directory], args.algorithm,
            options,
            ParallelOptions(
                args.parallel, args.chunk_size if args.chunk_size else CHUNK_SIZE
            )
        )
        logging.info("%d matches", len(matches))
        execute_actions(matches, args.on_equal)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



__author__ = 'Lene Preuss <lene.preuss@gmail.com>'

from functools import lru_cache, partial
from hashlib import sha256
from pathlib import Path
from subprocess import call  # noqa: S404
from typing import Dict, Tuple, List

from duplicate_images.function_types import ActionFunction, AlgorithmOptions, ComparisonFunction
from duplicate_images.histogram import compare_image_histograms
from duplicate_images.image_wrapper import ImageWrapper
from duplicate_images.image_hash import is_similar, IMAGE_HASH_ALGORITHM


@lru_cache(maxsize=None)
def get_size(file: Path) -> int:
    return file.stat().st_size


@lru_cache(maxsize=None)
def get_hash(file: Path) -> str:
    return sha256(file.open('rb').read()).hexdigest()


def compare_exactly(
        file: Path, other_file: Path, options: AlgorithmOptions  # pylint: disable=unused-argument
) -> bool:
    """Returns True if file and other_file are exactly exactly_equal"""
    return get_size(other_file) == get_size(file) and get_hash(file) == get_hash(other_file)


def compare_histograms(
        file: Path, other_file: Path, options: AlgorithmOptions
) -> bool:
    """Returns True if the histograms of file and other_file differ by
       less than rms_error"""
    aspect_fuzziness = options['aspect_fuzziness']
    rms_error = options['rms_error']
    try:
        return compare_image_histograms(
            ImageWrapper.create(file), ImageWrapper.create(other_file), aspect_fuzziness, rms_error
        )
    except (IOError, TypeError):
        return False


def compare_image_hash(
        algo: str, file: Path, other_file: Path,
        options: AlgorithmOptions  # pylint: disable=unused-argument
) -> bool:
    return is_similar(
        ImageWrapper.create(file), ImageWrapper.create(other_file),
        IMAGE_HASH_ALGORITHM[algo]
    )


def ascending_by_size(pair: Tuple[Path, Path]) -> List[Path]:
    return sorted(pair, key=lambda path: path.stat().st_size)


COMPARISON_METHODS: Dict[str, ComparisonFunction] = {
    'exact': compare_exactly,
    'histogram': compare_histograms,
}
for key in IMAGE_HASH_ALGORITHM:
    COMPARISON_METHODS[key] = partial(compare_image_hash, key)

ACTIONS_ON_EQUALITY: Dict[str, ActionFunction] = {
    'delete-first': lambda pair: pair[0].unlink(),
    'delete-second': lambda pair: pair[1].unlink(),
    'delete-bigger': lambda pair: ascending_by_size(pair)[-1].unlink(),
    'delete-smaller': lambda pair: ascending_by_size(pair)[0].unlink(),
    'eog': lambda pair: call(["eog"] + [str(pic) for pic in pair]),  # noqa: S603
    'xv': lambda pair: call(["xv", "-nolim"] + [str(pic) for pic in pair]),  # noqa: S603
    'print': lambda pair: print(pair[0], pair[1]),
    'none': lambda pair: None
}



__author__ = 'lene'

from argparse import ArgumentParser, Namespace

from duplicate_images.methods import COMPARISON_METHODS, ACTIONS_ON_EQUALITY


def parse_command_line() -> Namespace:
    parser = ArgumentParser(description="Find pairs of equal or similar images.")

    parser.add_argument(
        'root_directory', default='.', nargs='+',
        help="The root of the directory tree under which images are compared"
    )
    parser.add_argument(
        '--algorithm', choices=COMPARISON_METHODS.keys(),
        default='exact',
        help="Method used to determine if two images are considered equal"
    )
    parser.add_argument(
        '--on-equal', choices=ACTIONS_ON_EQUALITY.keys(),
        default='print', help="command to be run on each pair of images found to be equal"
    )
    parser.add_argument(
        '--parallel', action='store_true', help="Filter using all available cores (Experimental)"
    )
    parser.add_argument(
        '--chunk-size', type=int, help="Chunk size for parallelization"
    )
    parser.add_argument(
        '--fuzziness', '-f', default=0.001, type=float,
        help="Maximum deviation (RMS) of the histograms of two images still considered equal"
    )
    parser.add_argument(
        '--aspect-fuzziness', default=0.05, type=float,
        help="Maximum difference in aspect ratios of two images to compare more closely"
    )
    parser.add_argument(
        '--debug', action='store_true', help="Print lots of debugging info"
    )

    return parser.parse_args()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['duplicate_images']

package_data = \
{'': ['*']}

install_requires = \
['Wand>=0.6.5,<0.7.0', 'imagehash>=4.2.0,<5.0.0', 'pillow>=8.1.0,<9.0.0']

entry_points = \
{'console_scripts': ['find-dups = duplicate_images.duplicate:main']}

setup_kwargs = {
    'name': 'duplicate-images',
    'version': '0.3.1',
    'description': '',
    'long_description': "# Finding Duplicate Images\n\nFinds equal or similar images in a directory containing (many) image files.\n\n## Usage\n```shell\n$ pip install duplicate_images\n$ find-dups -h\n```\nto print the help screen. Or just\n```shell\n$ find-dups $IMAGE_ROOT \n```\nfor a test run.\n\n### Image comparison algorithms\n\nUse the `--algorithm` option to select how equal images are found.\n- `exact`: marks only binary exactly equal files as equal. This is by far the fastest, but most \n  restricted algorithm.\n- `histogram`: checks the images' color histograms for equality. Faster than the image hashing \n  algorithms, but tends to give a lot of false positives for images that are similar, but not equal.\n  Use the `--fuzziness` and `--aspect-fuzziness` options to fine-tune its behavior.\n- `ahash`, `colorhash`, `dhash`, `phash`, `whash`: five different image hashing algorithms. See \n  https://pypi.org/project/ImageHash for an introduction on image hashing and \n  https://tech.okcupid.com/evaluating-perceptual-image-hashes-okcupid for some gory details which\n  image hashing algorithm performs best in which situation. For a start I recommend using `ahash`, \n  and only evaluating the other algorithms if `ahash` does not perform satisfactorily in your use \n  case.\n\n### Actions for matching image pairs\n\nUse the `--on-equal` option to select what to do to pairs of equal images.\n- `delete-first`: deletes the first of the two files\n- `delete-second`: deletes the second of the two files\n- `delete-bigger`: deletes the file with the bigger size\n- `delete-smaller`: deletes the file with the smaller size\n- `eog`: launches the `eog` image viewer to compare the two files\n- `xv`: launches the `xv` image viewer to compare the two files\n- `print`: prints the two files\n- `none`: does nothing.\nThe default action is `print`.\n  \n### Parallel execution\n\nUse the `--parallel` option to utilize all free cores on your system. There is also the \n`--chunk-size` option to tune how many comparisons each thread should make in one go, but that \nshould hardly ever be advantageous to set explicitly. \n\n## Development notes\n\nNeeds Python3 and Pillow imaging library to run, additionally Wand for the test suite.\n\nUses Poetry for dependency management.\n\n### Installation\n\nFrom source:\n```shell\n$ git clone https://gitlab.com/lilacashes/DuplicateImages.git\n$ cd DuplicateImages\n$ pip3 install poetry\n$ poetry install\n```\n\n### Running\n\n```shell\n$ poetry run find-dups $PICTURE_DIR\n```\nor\n```shell\n$ poetry run find-dups -h\n```\nfor a list of all possible options.\n\n### Test suite\n\nRunning it all:\n```shell\n$ poetry run pytest\n$ poetry run mypy duplicate_images tests\n$ poetry run flake8\n$ poetry run pylint duplicate_images tests\n```\nor simply \n```shell\n$ .git_hooks/pre-push\n```\nSetting the test suite to be run before every push:\n```shell\n$ cd .git/hooks\n$ ln -s ../../.git_hooks/pre-push .\n```\n\n### Publishing\n\n```shell\n$ poetry config repositories.testpypi https://test.pypi.org/legacy/\n$ poetry build\n$ poetry publish --username $PYPI_USER --password $PYPI_PASSWORD --repository testpypi && \\\n  poetry publish --username $PYPI_USER --password $PYPI_PASSWORD\n```\n(obviously assuming that username and password are the same on PyPI and TestPyPI)\n### Profiling\n\n#### CPU time\nTo show the top functions by time spent, including called functions:\n```shell\n$ poetry run python -m cProfile -s tottime ./duplicate_images/duplicate.py \\ \n    --algorithm $ALGORITHM --action-equal none $IMAGE_DIR 2>&1 | head -n 15\n```\nor, to show the top functions by time spent in the function alone:\n```shell\n$ poetry run python -m cProfile -s cumtime ./duplicate_images/duplicate.py \\ \n    --algorithm $ALGORITHM --action-equal none $IMAGE_DIR 2>&1 | head -n 15\n```\n\n#### Memory usage\n```shell\n$ poetry run fil-profile run ./duplicate_images/duplicate.py \\\n    --algorithm $ALGORITHM --action-equal none $IMAGE_DIR 2>&1\n```\nThis will open a browser window showing the functions using the most memory (see \nhttps://pypi.org/project/filprofiler for more details).",
    'author': 'Lene Preuss',
    'author_email': 'lene.preuss@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/lilacashes/DuplicateImages',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.5,<4.0.0',
}


setup(**setup_kwargs)
